from typing import List, Dict
from dataclasses import dataclass, field
from ._helpers import build_if_set

KNOWN_COMPRESSION = ["auto", "off", "max"]


@dataclass
class Restic(object):
    image: str = "restic/restic:0.18.0"
    compression: str = "auto"
    no_snapshot_unless_changed: bool = True
    extra_args: List[str] = field(default_factory=list)

    def __post_init__(self) -> None:
        if self.compression not in KNOWN_COMPRESSION:
            raise ValueError(
                f"unknown compression `{self.compression}` the valid option are {', '.join(KNOWN_COMPRESSION)}"
            )


@dataclass
class BackupPath(object):
    default: str = "${HOME}/.sandock_vol_backup"
    # TODO: path maps to backup repo based on label


@dataclass
class Backup(object):
    restic: Restic = field(default_factory=Restic)
    path: BackupPath = field(default_factory=BackupPath)
    no_password: bool = False
    volume_labels: Dict[str, str] = field(default_factory=dict)
    volume_excludes: List[str] = field(default_factory=list)

    def __post_init__(self) -> None:
        build_if_set(self, attr="restic", cls=Restic)
        build_if_set(self, attr="path", cls=BackupPath)



import re
from re import Pattern as RegexPattern
from dataclasses import dataclass, field
from typing import List, Optional
from pathlib import Path
from ..shared import log, dict_merge, KV
from ._helpers import dot_config_finder, read_config


@dataclass
class Configuration(object):
    current_dir_conf: bool = True
    current_dir_conf_excludes: List[RegexPattern] = field(default_factory=list)  # type: ignore[type-arg]
    includes: List[str] = field(default_factory=list)

    def __post_init__(self) -> None:
        # convert to the list of regex patterns, intended as a list for next feature comes
        rgx_pattern_props = [
            "current_dir_conf_excludes",
        ]
        for rgx_prop in rgx_pattern_props:
            ptrns = []
            for val in getattr(self, rgx_prop):
                ptrns.append(re.compile(val))

            setattr(self, rgx_prop, ptrns)

    def filter_current_dir_conf(self, location: Path) -> Optional[str]:
        """
        determine whether it's allowed based on exclude pattern list
        """
        fpath = str(location)
        for current_dir_exlude in self.current_dir_conf_excludes:
            if current_dir_exlude.fullmatch(fpath):
                log.debug(
                    f"[config] current directory config `{fpath}` is excluded due match with the list (`{current_dir_exlude}`)"
                )
                return None

        return fpath

    @property
    def dir_conf(self) -> Optional[str]:
        """
        return the path of current directory conf if it's enabled and the file is exists
        """
        if not self.current_dir_conf:
            return None

        current_path = Path.cwd()
        dot_current = dot_config_finder(directory=current_path)
        return (
            self.filter_current_dir_conf(location=dot_current) if dot_current else None
        )

    def expand_configs(self) -> KV:
        """
        list of expanded config
        """
        tobe_merged: KV = {}

        # include list
        for inc in self.includes:
            log.debug(f"[config] read included file {inc}")
            tobe_merged = dict_merge(tobe_merged, read_config(path=inc))

        # dot configuration
        current_dir_config = self.dir_conf
        if current_dir_config:
            log.debug(f"[config] read current configuration dir {current_dir_config}")
            tobe_merged = dict_merge(tobe_merged, read_config(path=current_dir_config))

        return tobe_merged



import os
import json
import tempfile
import re
from datetime import datetime
from typing import List, Dict, Any, Optional, Callable, Tuple
from pathlib import Path
from .config import MainConfig
from .config.program import Program
from .config.image import ImageBuild, DEFAULT_DUMP_IMAGE_STORE
from .shared import log, run_shell, file_hash, ensure_home_dir_special_prefix
from .exceptions import SandboxExecution

VOL_LABEL_CREATED_BY = "created_by.sandock"
VOL_SANDBOX_DIR_SUFFIX_NAME = "sandock-dir-"
VOL_DIR_SUBS = "${VOL_DIR}"
CONTAINER_STATE_RUNNING = "running"
special_char_re = re.compile(r"\W+")


class SandboxExec(object):
    cfg: MainConfig
    program_name: str
    program: Program
    interactive: bool
    container_name: str

    def __init__(
        self, name: str, cfg: MainConfig, overrides: Dict[str, Any] = {}
    ) -> None:
        self.cfg = cfg
        self.name = name
        program = self.cfg.programs.get(name)
        if not program:
            raise SandboxExecution(f"`{name}` is not defined")

        # persist container cannot be renamed in preventing unexpected behaviour
        # (eg: need gc the stopped one with different name)
        if program.persist.enable and "name" in overrides:
            raise SandboxExecution("name of persist program cannot be overrided")

        hooks: List[Tuple[Callable[...], Any]] = []  # type: ignore[misc]
        # apply program's attribute overrides
        for k, v in overrides.items():
            if not hasattr(program, k):
                # it might be an internal/hook method
                method = f"hook_{k}"
                if hasattr(self, method):
                    log.debug(f"hook detected for method {method}")
                    hooks.append((getattr(self, method), v))
                    continue

                log.warning(f"program doesn't has property {k}")
                continue

            log.debug(f"overriding value {v} in property {k}")
            setattr(program, k, v)
        self.program = program

        # prevent if it's run on homedir, we don't want unintended breach except their aware
        if (
            self.program.sandbox_mount.enable
            and not self.program.allow_home_dir
            and self.current_dir == self.home_dir
        ):
            raise SandboxExecution(
                "cannot be ran on top of home directory when the program's sandbox mount is enabled"
            )

        self.container_name = self.generate_container_name()
        # run hooks if any
        for method_hook, arg in hooks:
            method_hook(arg)

    def hook_recreate_img(self, create: bool=False) -> None:
        """
        register for pre-exec cmd to delete image run the related container
        """
        if not create:
            return

        log.debug("[hook] registring for image deletion: {self.program.image}")
        self.program.pre_exec_cmds.insert(0, " ".join([
            self.docker_bin, "image", "rm", self.program.image
        ]))

    @property
    def docker_bin(self) -> str:
        return self.cfg.execution.docker_bin

    @property
    def current_timestamp(self) -> float:
        return datetime.now().timestamp()

    def generate_container_name(self) -> str:
        if self.program.name:
            return self.program.name

        generated_name = f"{self.cfg.execution.container_name_prefix}{self.name}"
        # don't give timestampt if it's persist
        if not self.program.persist.enable:
            generated_name += f"-{self.current_timestamp}"
        return generated_name

    @property
    def home_dir(self) -> Optional[str]:
        return os.environ.get("HOME")

    @property
    def current_dir(self) -> str:
        return os.getcwd()

    @property
    def exec_path(self) -> str:
        """
        grab the alias first
        """
        pe = self.program.exec
        return self.program.aliases.get(pe, pe)

    def run_container_cmd(self) -> List[str]:
        """
        docker run command builder
        """
        command = [
            self.docker_bin,
            "run",
            "--entrypoint",
            self.exec_path,
            "--name",
            self.container_name,
        ]

        if not self.program.persist.enable:
            command.append("--rm")

        if self.program.interactive:
            command.append("-it")

        if self.program.platform:
            command.extend(["--platform", self.program.platform])

        if self.program.hostname:
            command.extend(["--hostname", self.program.hostname])

        if self.program.network:
            command.extend(["--network", self.program.network])

        user = self.program.user
        if user:
            uid = self.current_uid if user.keep_id else user.uid
            gid = self.current_gid if user.keep_id else user.gid
            command.extend(["-u", f"{uid}:{gid}"])

        workdir = self.program.workdir
        volumes = self.program.volumes
        if self.program.sandbox_mount.enable:
            # when sandbox mount enable, set the default workdir to it
            workdir = self.program.sandbox_mount.current_dir_mount
            sandbox_mount = (
                f"{self.current_dir}:{self.program.sandbox_mount.current_dir_mount}"
            )
            if self.program.sandbox_mount.read_only:
                sandbox_mount += ":ro"

            volumes.append(sandbox_mount)

        for v in volumes:
            if VOL_DIR_SUBS in v:
                log.debug("[vol] volume dir identified, subs the var")
                v = v.replace(
                    VOL_DIR_SUBS, special_char_re.sub("_", self.current_dir[1:])
                )
            command.extend(["-v", v])

        if workdir:
            command.extend(["--workdir", workdir])

        command.extend([f"-e {k}='{v}'" for k, v in self.program.env.items()])
        command.extend([f"-p {p}" for p in self.program.ports])
        command.extend([f"--cap-add={c}" for c in self.program.cap_add])
        command.extend([f"--cap-drop={c}" for c in self.program.cap_drop])
        command.extend(self.program.extra_run_args)
        command.append(self.program.image)

        return command

    @property
    def current_uid(self) -> int:
        return os.getuid()

    @property
    def current_gid(self) -> int:
        return os.getgid()

    def ensure_volume(self, name: str) -> None:
        """
        ensure the volume created as expected based on configuration if exists
        """
        log.debug(f"[vol] check volume by name {name} in configuration")
        vol = self.cfg.volumes.get(name)
        if not vol:
            log.debug(f"[vol] {name} not exists in config")
            return

        inspect_vol = run_shell(
            f"{self.docker_bin} volume inspect {name}", check_err=False
        )
        # vol exists
        if inspect_vol.returncode == 0:
            vol_info = json.loads(inspect_vol.stdout)
            log.debug(f"[vol] exists, info ~> {vol_info}")
            return

        log.info(f"[vol] creating vol {name}")
        # mandatory label
        if VOL_LABEL_CREATED_BY not in vol.labels:
            vol.labels[VOL_LABEL_CREATED_BY] = "true"

        vol_ops = " ".join([f"--opt {k}={v}" for k, v in vol.driver_opts.items()])
        vol_labels = " ".join([f"--label {k}='{v}'" for k, v in vol.labels.items()])
        vol_create_cmd = (
            f"{self.docker_bin} volume create "
            f"--driver={vol.driver} {vol_ops} {vol_labels}"
            f" {name}"
        )
        run_shell(vol_create_cmd, capture_output=False)

    def ensure_network(self) -> None:
        net_name = self.program.network
        if not net_name:
            return

        log.debug(f"[net] ensuring network by name `{net_name}` in configuration")
        network = self.cfg.networks.get(net_name)
        if not network:
            log.debug(
                f"[net] network by name {net_name} is not managed in configuration"
            )
            return

        inspect_net = run_shell(
            f"{self.docker_bin} network inspect {net_name}", check_err=False
        )
        if inspect_net.returncode == 0:
            net_info = json.loads(inspect_net.stdout)
            log.debug(f"[net] exists, info ~> {net_info}")
            return

        log.info(f"[net] creating net {net_name}")

        net_params = " ".join([f"--{k}={v}" for k, v in network.params.items()])
        net_driver_opts = " ".join(
            [f"--opt {k}={v}" for k, v in network.driver_opts.items()]
        )
        net_create_cmd = (
            f"{self.docker_bin} network create "
            f"--driver={network.driver} {net_params} {net_driver_opts}"
            f"{net_name}"
        )
        run_shell(net_create_cmd, capture_output=False)

    def custom_image_dockerfile_store(
        self, path: str, image_name: str, build: ImageBuild
    ) -> Path:
        docker_file_hash = file_hash(fpath=path, max_chars=20)
        return Path(
            (
                ensure_home_dir_special_prefix(path=build.dump.store)
                .replace("${image}", image_name)
                .replace("${hash}", docker_file_hash)
                .replace(
                    "${platform}",
                    (
                        ""
                        if not self.program.platform
                        else self.program.platform.replace("/", "_")
                    ),
                )
            )
        )

    def ensure_custom_image(self, image_name: Optional[str] = None) -> None:
        """
        if it's intended to use custom image
        """
        image_name = image_name or self.program.image
        build_image = self.cfg.images.get(image_name) or self.program.build
        if not build_image:
            return

        # recursively check for any dependent before build
        if build_image.depends_on:
            log.debug(f"found image depends on {build_image.depends_on}")
            self.ensure_custom_image(image_name=build_image.depends_on)

        inspect_img = run_shell(
            f"{self.docker_bin} image inspect {image_name}", check_err=False
        )

        if inspect_img.returncode == 0:
            img_info = json.loads(inspect_img.stdout)
            log.debug(f"[img] exists, info ~> {img_info}")
            return

        log.info(f"[img] creating image {image_name}")
        working_dir = build_image.context
        if not working_dir:
            # if context not set then create the temporary one
            log.debug("[img] context was not set, create the temporary dir")
            working_dir = tempfile.mkdtemp()
        else:
            working_dir = ensure_home_dir_special_prefix(path=working_dir)

        docker_file_path = build_image.dockerFile
        if not docker_file_path and build_image.dockerfile_inline:
            log.debug(
                "[img] docker file not defined, create the temporary one from `dockerfile_inline`"
            )
            docker_file_path = tempfile.mktemp()
            with open(docker_file_path, "w") as fh:
                fh.write(
                    build_image.dockerfile_inline.replace(
                        "{CURRENT_GID}", str(self.current_gid)
                    ).replace("{CURRENT_UID}", str(self.current_uid))
                )

        docker_file_path = ensure_home_dir_special_prefix(path=docker_file_path)  # type: ignore[arg-type]

        # if dumped image enabled and found then just load from it and skip image building
        dump_store_path = self.custom_image_dockerfile_store(
            path=docker_file_path, image_name=image_name, build=build_image
        )
        if build_image.dump.enable and dump_store_path.exists():
            log.info(
                f"[img] image dump is enable and cache image found, will restore the image from file `{dump_store_path}`"
            )
            image_restore_cmd = f"{self.docker_bin} image load -i {dump_store_path}"
            run_shell(image_restore_cmd, capture_output=False)
            return

        img_create_cmd = f"{self.docker_bin} build -t {image_name}"
        if docker_file_path:
            img_create_cmd += f" -f {docker_file_path}"

        for k_arg, v_arg in build_image.args.items():
            img_create_cmd += f' --build-arg {k_arg}="{v_arg}"'

        if build_image.extra_build_args:
            img_create_cmd += f' {" ".join(build_image.extra_build_args)}'

        if self.program.platform:
            img_create_cmd += f" --platform={self.program.platform}"

        img_create_cmd += f" {working_dir}"
        run_shell(img_create_cmd, capture_output=False)

        # dump image if it's enabled
        if build_image.dump.enable and not dump_store_path.exists():
            # clean up previous one
            if dump_store_path.parent.exists() and build_image.dump.cleanup_prev:
                log.info("[img] cleanup previous dumped image")
                base_store = os.path.basename(build_image.dump.store)

                # if using other than known format then just ignore it
                if base_store == os.path.basename(DEFAULT_DUMP_IMAGE_STORE):
                    for prev_img in dump_store_path.parent.glob(f"{image_name}*.tar"):
                        log.info(f"[img] cleanup previous cached image {prev_img}")
                        prev_img.unlink()
                else:
                    log.warning(
                        f"[img] it's using non standard pattern ({base_store}), cannot run cleanup operation"
                    )

            # ensure the parent directory is exists
            dump_store_path.parent.mkdir(parents=True, exist_ok=True)
            log.info(f"[img] save dumped image into {dump_store_path}")
            image_dump_cmd = (
                f"{self.docker_bin} image save {image_name} --output {dump_store_path}"
            )
            run_shell(image_dump_cmd, capture_output=False)

        # garbage collect
        if (
            build_image.dockerfile_inline
            and docker_file_path is not None
            and os.path.isfile(docker_file_path)
        ):
            os.remove(docker_file_path)

    @property
    def attach_container(self) -> bool:
        """
        determine whether attaching to existsing container ?
        """
        if not self.program.persist.enable:
            return False

        cmd = f"{self.docker_bin} container inspect {self.container_name}"
        inspect_result = run_shell(cmd, check_err=False)
        if inspect_result.returncode != 0:
            # it might be the uncreated container
            err_msg = str(inspect_result.stderr)
            if "no such container" in err_msg.lower():
                return False

            # the known message if container not found, other than that just raise it
            raise SandboxExecution(f"error during check container status: {err_msg}")

        container_info = json.loads(inspect_result.stdout)
        if len(container_info) < 1:
            log.warning(
                f"empty container info status during inspection ~> {container_info}"
            )
            return False

        is_running: bool = (
            container_info[0].get("State", {}).get("Status") == CONTAINER_STATE_RUNNING
        )
        if not is_running and self.program.persist.auto_start:
            log.warning("persist container is down, starting container")
            run_shell(
                f"{self.docker_bin} container start {self.container_name}",
                capture_output=False,
            )

            # just attach to it after start
            return True

        return is_running

    def exec_container_cmd(self) -> List[str]:
        cmd = [self.docker_bin, "exec"]

        if self.program.interactive:
            cmd.append("-it")

        cmd.extend([self.container_name, self.exec_path])
        return cmd

    def do(self, args: List[str] = []) -> None:
        """
        main flows
        """
        for pre_exec in self.program.pre_exec_cmds:
            log.debug(f"[pre_exec] running shell command ~> {pre_exec}")
            run_shell(pre_exec)

        # ensure container dependencies are created
        self.ensure_custom_image()
        self.ensure_network()
        for extra_mount in self.program.volumes:
            vol = extra_mount.split(":")[0]
            self.ensure_volume(name=vol)

        run_container_cmd = (
            self.exec_container_cmd()
            if self.attach_container
            else self.run_container_cmd()
        )
        run_container_cmd.extend(args)

        log.debug(f"starting container cmd ~> {run_container_cmd}")
        run_shell(command=" ".join(run_container_cmd), capture_output=False)



import json
from typing import Any, Optional
from ..shared import KV, ensure_home_dir_special_prefix, log
from ..exceptions import SandboxExecConfig
from collections import OrderedDict
from pathlib import Path

DOT_CONFIG = ".sandock"


def json_decoder(content: str) -> KV:
    """
    convert json content to dictionary
    """
    return json.loads(content)  # type: ignore[no-any-return]


def yaml_decoder(content: str) -> KV:
    """
    convert json content to dictionary
    """

    try:
        import yaml
    except ImportError:
        raise SandboxExecConfig(
            "yaml parser module not installed, try to install by following command:\npip install 'sandock[yml-config]'\nor manual module install:\npip install pyyaml"
        )
    raw = yaml.safe_load(content)

    # remove all that startswith x-*
    return {k: v for k, v in raw.items() if not k.startswith("x-")}


# the list of supported formatted configuration, default to json decoder
# ordering is critical here, it also determines what will be look up first
CONFIG_FORMAT_DECODER_MAPS = OrderedDict()
CONFIG_FORMAT_DECODER_MAPS[".yml"] = yaml_decoder
CONFIG_FORMAT_DECODER_MAPS[".yaml"] = yaml_decoder
CONFIG_FORMAT_DECODER_MAPS[".json"] = json_decoder


def read_config(path: str) -> KV:
    """
    read configuration as dict/kv based on it's decoder
    """
    conf_format = Path(path).suffix
    decoder = CONFIG_FORMAT_DECODER_MAPS.get(conf_format, json_decoder)

    # ensure the proper home dir pattern is set
    path = ensure_home_dir_special_prefix(path=path)

    with open(path, "r") as fh:
        return decoder(content=fh.read())


def dot_config_finder(directory: Path) -> Optional[Path]:
    """
    the logic behind dot configuration file based on given directory path
    """

    dot_config = directory / DOT_CONFIG
    for dot_format in CONFIG_FORMAT_DECODER_MAPS.keys():
        dot_current_format = dot_config.with_suffix(dot_format)
        log.debug(
            f"[config] dot config finder: searching by format in path {dot_current_format}"
        )
        if dot_current_format.exists():
            return dot_current_format

    log.debug(
        f"[config] dot config finder: searching without format in path {dot_config}"
    )

    return dot_config if dot_config.exists() else None


def build_if_set(o: object, attr: str, cls: Any) -> None:
    """
    helper, if the property is set with dictionary then build it as object with given class
    """
    prop = getattr(o, attr)
    if prop and isinstance(prop, dict):
        setattr(o, attr, cls(**prop))



import logging
import subprocess
import os
import re
from hashlib import sha256
from typing import Any, Union, Dict, List, Optional


KV = Dict[str, Any]
CONFIG_PATH_ENV = "SNDK_CFG"
SANDBOX_DEBUG_ENV = "SNDK_DEBUG"


class LogColorFormatter(logging.Formatter):
    # ref: https://alexandra-zaharia.github.io/posts/make-your-own-custom-color-formatter-with-python-logging/
    grey = "\x1b[38;21m"
    blue = "\x1b[38;5;39m"
    yellow = "\x1b[38;5;226m"
    red = "\x1b[38;5;196m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"

    def __init__(self, fmt: str) -> None:
        super().__init__()
        self.fmt = fmt
        self.FORMATS = {
            logging.DEBUG: self.grey + self.fmt + self.reset,
            logging.INFO: self.blue + self.fmt + self.reset,
            logging.WARNING: self.yellow + self.fmt + self.reset,
            logging.ERROR: self.red + self.fmt + self.reset,
            logging.CRITICAL: self.bold_red + self.fmt + self.reset,
        }

    def format(self, record: Any) -> str:
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def init_logger(name: str, lvl: Union[int, str] = logging.DEBUG) -> logging.Logger:
    pattern = "%(asctime)s | %(levelname)3s | %(message)s"
    logger = logging.getLogger(name)
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(LogColorFormatter(pattern))
    logger.setLevel(lvl)
    logger.addHandler(handler)

    return logger


log = init_logger(name="sandbox-exec", lvl=logging.INFO)


def run_shell(  # type: ignore[no-untyped-def]
    command: Union[str, List[str]], check_err: bool = True, **cmd_args
) -> subprocess.CompletedProcess:  # type: ignore[type-arg]
    """
    wrapper of shell command execution
    """
    cmd_args = (
        dict(
            shell=True,
            capture_output=True,
            text=True,
        )
        | cmd_args
    )
    if isinstance(command, list):
        command = " ".join(command)

    log.debug(f"shell cmd: {command}, check_err: {check_err}, cmd_args: {cmd_args}")
    call_cmd = subprocess.run(command, **cmd_args)
    if check_err and call_cmd.returncode != 0:
        msg = f"error in executing command: {command}"
        if call_cmd.stderr:
            msg += f", stderr: {call_cmd.stderr}"
        if call_cmd.stdout:
            msg += f", stdout: {call_cmd.stdout}"
        raise subprocess.CalledProcessError(
            cmd=msg,
            returncode=call_cmd.returncode,
        )
    return call_cmd


def dict_merge(dict1: KV, dict2: KV) -> KV:
    """
    deep merge between dict1 and dicgt2
    """
    result = dict1.copy()
    for key, value in dict2.items():
        if key in result and isinstance(result[key], dict) and isinstance(value, dict):
            result[key] = dict_merge(result[key], value)
        else:
            result[key] = value

    return result


home_dir_re = re.compile(r"^(~|\$HOME|\$\{HOME\})")


def ensure_home_dir_special_prefix(path: str) -> str:
    """
    ensure the path that begins with ~, $HOME, ${HOME}
    are converted with the real home directory
    """
    return home_dir_re.sub(os.environ["HOME"], path)


def file_hash(fpath: str, max_chars: Optional[int] = None) -> str:
    """
    calculate SHA256 of given text, default will return as 64 chars but it can be limited
    """
    with open(fpath, "r") as fh:
        hex_digest = sha256(fh.read().encode("utf-8")).hexdigest()

        return hex_digest[:max_chars] if max_chars is not None else hex_digest



import os
import sys
import logging
import subprocess
from typing import List, Tuple, Dict, Any, Optional
from argparse import ArgumentParser, Namespace, REMAINDER, ArgumentTypeError
from importlib.metadata import metadata
from .config import MainConfig, load_config_file, main_config_finder
from .shared import log, SANDBOX_DEBUG_ENV, CONFIG_PATH_ENV, run_shell
from .sandbox import SandboxExec
from .volume import VolumeMgr
from .exceptions import SandboxBaseException, SandboxExecConfig, SandboxVolumeExec
from ._version import __version__, __build_hash__


def parse_arg_key_value(s: str) -> Tuple[str, str]:
    if "=" not in s:
        raise ArgumentTypeError(f"Invalid format: '{s}', expected KEY=VALUE")

    key, value = s.split("=", 1)
    return key, value


class BaseCommand(object):
    args: Namespace
    config: MainConfig
    description: str = ""

    def __init__(self, args: Namespace):
        self.args = args
        self.config = self._read_config()

    def output(self, msg: str) -> None:
        """
        a wrapper for any output to user (stdout), also as the easier way in test side
        """
        sys.stdout.write(f"{msg}\n")

    def override_arg(self, dashed: bool = True, name: str = "") -> str:
        """
        generate the override argument
        """
        prefix = self.config.execution.property_override_prefix_arg
        prefix = f"--{prefix}" if dashed else prefix
        return f"{prefix}{name}"

    @property
    def config_path(self) -> Optional[str]:
        """
        get main configuration path
        """
        return main_config_finder(explicit_mention=self.args.config)

    def _read_config(self) -> MainConfig:
        """
        read main configuration path file as conf object
        """
        config_path = self.config_path
        if config_path is None:
            raise SandboxExecConfig("no main configuration can be read")

        if not os.path.isfile(config_path):
            raise SandboxExecConfig(
                f"main configuration is not found (`{config_path}`)"
            )

        return load_config_file(path=config_path)

    @staticmethod
    def register_arguments(parser: ArgumentParser) -> None:
        pass

    def main(self) -> None:
        """
        main execution, should be extended
        """


class CmdList(BaseCommand):
    description = "list available sandboxed program, the name also added with a prefix name if configured"

    def main(self) -> None:
        for prog_name in self.config.programs.keys():
            self.output(prog_name)


class CmdAlias(BaseCommand):
    description = "print the list of alias as a shortcut to ran the programs, this should be added in shell profile configuration"

    @property
    def executor(self) -> str:
        """
        return executor name
        """
        return sys.argv[0]

    @staticmethod
    def register_arguments(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--expand", help="include with aliases", action="store_true", default=False
        )

        parser.add_argument(
            "program_args",
            nargs=REMAINDER,
            help="program argument that will be forwarded",
        )

    def main(self) -> None:
        for prog_name, prog_cfg in self.config.programs.items():
            alias_key = f"{self.config.execution.alias_program_prefix}{prog_name}"
            alias_value = f"{self.executor} run {prog_name}"
            main_line = f'alias {alias_key}="{alias_value}"'
            self.output(main_line)
            if not self.args.expand:
                continue

            for alias_cmd in prog_cfg.aliases.keys():
                alias_exec_line = f'alias {alias_key}-{alias_cmd}="{alias_value} {self.override_arg(name="exec")}={alias_cmd}"'
                self.output(alias_exec_line)


class CmdVolume(BaseCommand):
    description = "manage container volumes"

    @staticmethod
    def register_arguments(parser: ArgumentParser) -> None:
        subs = parser.add_subparsers(
            title="volume action", dest="volume_action", required=True
        )

        # list
        subs.add_parser("list", help="list all volume that created by sandock")

        # backup subcmd
        backup_sub = subs.add_parser("backup", help="backup related command")
        # the default action is to create a backup
        backup_sub.add_argument(
            "-a",
            "--all",
            help="backup all volumes based mentioned labels in configuration",
            action="store_true",
            default=False,
        )
        backup_sub.add_argument(
            "--target", help="specific volume name that will be set as target backup"
        )
        backup_sub.add_argument(
            "-e", "--exclude", action="append", help="explicit exclude volume to backup"
        )
        backup_parser = backup_sub.add_subparsers(dest="backup_action")

        # backup - snapshot
        snapshot_sub = backup_parser.add_parser(
            "snapshot",
            help="show all existing backup snapshot, by default it's only shown the latest one",
        )
        snapshot_sub.add_argument(
            "--all",
            help="show all snapshots",
            action="store_true",
            default=False,
        )
        snapshot_sub.add_argument(
            "--vol",
            help="only show snapshot the specific volume backup",
            default=None,
        )

        # backup - restore
        restore_sub = backup_parser.add_parser(
            "restore", help="backup - volume restore command"
        )
        restore_sub.add_argument(
            "-i",
            "--snapshot-id",
            help="snapshot ID that will be restored",
            required=True,
        )
        restore_sub.add_argument(
            "--vol",
            help="target volume name that will be restored, if not exists then it will create the new one",
            required=True,
        )
        restore_sub.add_argument(
            "-f",
            "--force",
            help="force execution even the target volume is already exists, will be ran overrides action",
            action="store_true",
            default=False,
        )
        restore_sub.add_argument(
            "--exclude", action="append", help="exclude path to be restored"
        )
        restore_sub.add_argument(
            "--overwrite",
            help="set overwrite behaviour",
            default="never",
            choices=["always", "if-changed", "if-newer", "never"],
        )

        # backup - restic
        restic_sub = backup_parser.add_parser(
            "restic",
            help="backup - restic, direct restic command execution. use with cautions !!!",
        )
        restic_sub.add_argument(
            "--extra-run-args",
            help="additional docker run arguments, eg: mounting the volume etc.",
        )
        restic_sub.add_argument(
            "restic_params", nargs=REMAINDER, help="the forwarded arguments/params"
        )

    def run_backup(self, vol_mgr: VolumeMgr) -> None:
        # snapshot handler
        if self.args.backup_action == "snapshot":
            datetime_format = "%Y-%m-%d %H:%M:%S"
            for snapshot in vol_mgr.backup.snapshot_list(
                specific_volname=self.args.vol, show_all=self.args.all
            ):
                self.output(
                    f"id: {snapshot.id} - date: {snapshot.backup_time.strftime(datetime_format)} (UTC) - size: {snapshot.size} - vol: {snapshot.vol_name}"
                )

            return

        # restore handler
        if self.args.backup_action == "restore":
            backup_mgr = vol_mgr.backup
            snapshot = backup_mgr.get_snapshot_by(id=self.args.snapshot_id)
            backup_mgr.restore(
                snapshot=snapshot,
                target_volume=self.args.vol,
                force=self.args.force,
                excludes=self.args.exclude or [],
                overwrite=self.args.overwrite,
            )
            return

        # direct restic handler
        if self.args.backup_action == "restic":
            docker_params = []
            if self.args.extra_run_args:
                docker_params.append(self.args.extra_run_args)

            run_shell(
                command=vol_mgr.backup.restic_run_cmd(
                    extra_docker_params=docker_params,
                    restic_args=self.args.restic_params,
                ),
                capture_output=False,
            )
            return

        # create backup handler
        if self.args.target and self.args.all:
            raise SandboxVolumeExec(
                "cannot combine specific target with all volume option"
            )

        if not self.args.target and not self.args.all:
            raise SandboxVolumeExec("you must set explicitly volume backup target")

        targets = []
        if self.args.target:
            if not vol_mgr.vol_exists(name=self.args.target):
                raise SandboxVolumeExec(
                    f"volume by name `{self.args.target}` is not exists"
                )

            targets = [self.args.target]
        # if not specific then all volumes as configured
        else:
            vol_backup_labels = self.config.backup.volume_labels
            if not vol_backup_labels:
                raise SandboxVolumeExec(
                    "empty volume label filter for backup target, set it on .backup.volume.labels"
                )

            targets = [
                vol_info["Name"]
                for vol_info in vol_mgr.volume_list(label_filters=vol_backup_labels)
            ]

        if not targets:
            log.warning("no any volume backup target available")
            return

        vol_mgr.backup.create(targets=targets, excludes=self.args.exclude or [])

    def main(self) -> None:
        vol_mgr = VolumeMgr(cfg=self.config)

        sub_cmd = self.args.volume_action
        if sub_cmd == "list":
            for vol_info in vol_mgr.created_by_sandock:
                log.debug(f"volume info ~> {vol_info}")
                self.output(vol_info["Name"])

        elif sub_cmd == "backup":
            self.run_backup(vol_mgr=vol_mgr)


class CmdRun(BaseCommand):
    description = "run program"

    @property
    def overrides_args(self) -> ArgumentParser:
        """
        list of injected additional parameters for running docker run args
        """
        oparser = ArgumentParser(description="overriding parameters")

        oparser.add_argument(
            self.override_arg(name="name"), default=None, help="override container name"
        )

        oparser.add_argument(
            self.override_arg(name="exec"), default=None, help="override exec"
        )

        oparser.add_argument(
            self.override_arg(name="hostname"), default=None, help="override hostname"
        )

        oparser.add_argument(
            self.override_arg(name="network"), default=None, help="override network"
        )

        oparser.add_argument(
            self.override_arg(name="allow-home-dir"),
            action="store_true",
            default=False,
            help="override allow home directory mount on auto mount current directory",
        )

        oparser.add_argument(
            self.override_arg(name="env"),
            action="append",
            type=parse_arg_key_value,
            help="set environment in KEY=VALUE format",
        )

        oparser.add_argument(
            self.override_arg(name="ports"),
            action="append",
            help="publish container ports",
        )

        oparser.add_argument(
            self.override_arg(name="recreate-img"),
            action="store_true",
            default=False,
            help="recreate the used container image",
        )

        oparser.add_argument(
            self.override_arg(name="help"),
            action="store_true",
            default=False,
            help="show help",
        )

        return oparser

    @staticmethod
    def register_arguments(parser: ArgumentParser) -> None:
        parser.add_argument(
            "program",
        )

        parser.add_argument(
            "program_args",
            nargs=REMAINDER,
            help="arguments that will be forwarded, excluded for the override args",
        )

    def override_properties(self, args: List[str]) -> Dict[str, Any]:
        """
        convert the override argument to Program's property
        """
        result = {}
        kv_args = ["env"]
        ov_args = self.overrides_args
        for k, v in vars(ov_args.parse_args(args)).items():
            if v is None:
                continue
            arg_name = k.replace(self.override_arg(dashed=False).replace("-", "_"), "")
            if arg_name == "help":
                if v is True:
                    ov_args.print_help()
                    sys.exit(0)

                continue

            # convert to dict
            if arg_name in kv_args:
                v = dict(v or [])

            result[arg_name] = v
        return result

    @property
    def remainder_args(self) -> Tuple[List[str], Dict[str, str]]:
        """
        capture argument that will be forwarded to program and read for sandbox-exec
        """
        program_args = []
        snbx_args = []
        for remainder in self.args.program_args:
            if remainder.startswith(self.override_arg()):
                snbx_args.append(remainder)
                continue

            program_args.append(remainder)

        return (program_args, self.override_properties(args=snbx_args))

    def main(self) -> None:
        program_args, overrides = self.remainder_args
        log.debug(f"overrides args ~> {overrides}")
        snbx = SandboxExec(name=self.args.program, cfg=self.config, overrides=overrides)
        snbx.do(args=program_args)


def reraise_if_debug(e: Exception) -> None:
    """
    raise the exception if currently in debug mode
    """
    if log.level == logging.DEBUG:
        raise e


def main(args: Optional[List[str]] = None) -> None:
    meta = metadata("sandock")
    cmds = dict(list=CmdList, alias=CmdAlias, run=CmdRun, volume=CmdVolume)
    parser = ArgumentParser(
        description="A wrapper in running command inside container sandboxed environment",
        epilog=f"Author: {meta['author']} <{meta['author-email']}>",
    )
    parser.add_argument(
        "-c",
        "--config",
        help=f"path of configuration file, this can be overrided by env name `{CONFIG_PATH_ENV}`",
        default=None,
    )
    parser.add_argument(
        "-d",
        "--debug",
        help=f"enable debug mode, can be configured from env var `{SANDBOX_DEBUG_ENV}` by value `true`",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {__version__} (build {__build_hash__})",
    )
    subs = parser.add_subparsers(title="commands", dest="subparser")
    # register sub-commands
    for sub_name, sub_cls in cmds.items():
        sub = subs.add_parser(sub_name, description=sub_cls.description)
        sub_cls.register_arguments(parser=sub)
    parsed_args = parser.parse_args(args=args)

    if parsed_args.debug or os.environ.get(SANDBOX_DEBUG_ENV) == "true":
        log.setLevel(logging.DEBUG)

    # only registered subcommand can be executed
    exec_cls = cmds.get(parsed_args.subparser)
    if not exec_cls:
        return parser.print_help()

    # remove the python stack trace noise on non 0 exit, except in debug mode
    try:
        exec_cls(args=parsed_args).main()
    except subprocess.CalledProcessError as e:
        reraise_if_debug(e=e)

        log.error(f"exit code {e.returncode}, see the details in debug mode")
        sys.exit(e.returncode)

    except SandboxBaseException as e:
        reraise_if_debug(e=e)

        log.error(f"{e}, see the details in debug mode")
        sys.exit(1)



class SandboxBaseException(Exception):
    pass


class SandboxExecConfig(SandboxBaseException):
    """
    errors related to configuration
    """


class SandboxExecution(SandboxBaseException):
    pass


class SandboxVolumeExec(SandboxBaseException):
    pass


class SandboxVolumeNotFound(SandboxBaseException):
    pass



from dataclasses import dataclass, field
from typing import List, Optional, Dict
from ._helpers import build_if_set

DEFAULT_DUMP_IMAGE_STORE = (
    "${HOME}/.sandock_dump_images/${image}:${platform}${hash}.tar"
)


@dataclass
class ImageDump(object):
    enable: bool = False
    cleanup_prev: bool = True
    store: str = DEFAULT_DUMP_IMAGE_STORE


@dataclass
class ImageBuild(object):
    context: Optional[str] = None
    dockerfile_inline: Optional[str] = None
    dockerFile: Optional[str] = None
    depends_on: Optional[str] = None
    args: Dict[str, str] = field(default_factory=dict)
    extra_build_args: List[str] = field(default_factory=list)
    dump: ImageDump = field(default_factory=ImageDump)

    def __post_init__(self) -> None:
        if self.dockerfile_inline and self.dockerFile:
            raise ValueError("cannot set `dockerfile_inline` and `dockerFile` together")

        build_if_set(self, attr="dump", cls=ImageDump)



from dataclasses import dataclass, field
from typing import List, Optional, Dict
from .image import ImageBuild
from ._helpers import build_if_set


@dataclass
class SandboxMount(object):
    enable: bool = True
    read_only: bool = False
    current_dir_mount: str = "/sandbox"


@dataclass
class PersistContainer(object):
    enable: bool = False
    auto_start: bool = True


@dataclass
class ContainerUser(object):
    uid: int = 0
    gid: int = 0
    # keep both uid and gid
    keep_id: bool = False

    def __post_init__(self) -> None:
        if self.keep_id and self.uid != 0:
            raise ValueError(
                "cannot enabled on `keep_id` and set custom on `uid` in same time"
            )

        if self.keep_id and self.gid != 0:
            raise ValueError(
                "cannot enabled on `keep_id` and set custom on `gid` in same time"
            )


@dataclass
class Program(object):
    image: str
    exec: str
    interactive: bool = True
    allow_home_dir: bool = False
    name: Optional[str] = None
    network: Optional[str] = None
    hostname: Optional[str] = None
    build: Optional[ImageBuild] = None
    user: Optional[ContainerUser] = None
    workdir: Optional[str] = None
    platform: Optional[str] = None
    persist: PersistContainer = field(default_factory=PersistContainer)
    sandbox_mount: SandboxMount = field(default_factory=SandboxMount)
    env: Dict[str, str] = field(default_factory=dict)
    volumes: List[str] = field(default_factory=list)
    ports: List[str] = field(default_factory=list)
    cap_add: List[str] = field(default_factory=list)
    cap_drop: List[str] = field(default_factory=list)
    aliases: Dict[str, str] = field(default_factory=dict)
    extra_run_args: List[str] = field(default_factory=list)
    pre_exec_cmds: List[str] = field(default_factory=list)

    def __post_init__(self) -> None:
        build_if_set(self, attr="build", cls=ImageBuild)
        build_if_set(self, attr="user", cls=ContainerUser)
        build_if_set(self, attr="persist", cls=PersistContainer)
        build_if_set(self, attr="sandbox_mount", cls=SandboxMount)

        if self.sandbox_mount.enable and self.workdir:
            raise ValueError(
                "cannot use workdir with enabled sandbox mount in the same time"
            )



import json
import re
from getpass import getpass
from pathlib import Path
from dataclasses import dataclass
from typing import List, Optional, Union
from datetime import datetime
from .exceptions import SandboxVolumeExec, SandboxVolumeNotFound
from .config import MainConfig
from .config.backup import Backup as BackupCfg
from .sandbox import VOL_LABEL_CREATED_BY
from .shared import run_shell, log, KV, ensure_home_dir_special_prefix

BACKUP_MOUNT = "/backup_repo"
RESTORE_MOUNT_PREFIX = "/restore_vol_"
SOURCE_MOUNT_PREFIX = "/source_vol_"
BACKUP_IGNORE_REF_FILENAME = ".sandock_backup_ignore"

source_mount_prefix_re = re.compile(rf"^{SOURCE_MOUNT_PREFIX}")


def bytes_to_readable(num_bytes: float, decimal_places: int = 2) -> str:
    # convert bytes to humand readable
    units = ["B", "KB", "MB", "GB", "TB", "PB", "EB"]
    index = 0
    while num_bytes >= 1024 and index < len(units) - 1:
        num_bytes /= 1024.0
        index += 1

    return f"{num_bytes:.{decimal_places}f} {units[index]}"


@dataclass
class BackupSnapshot(object):
    id: str
    backup_time: datetime
    path: str
    bytes: int

    @property
    def vol_name(self) -> str:
        return source_mount_prefix_re.sub("", self.path)

    @property
    def size(self) -> str:
        # human readable size
        return bytes_to_readable(num_bytes=self.bytes)

    @classmethod
    def from_raw(self, data: KV) -> "BackupSnapshot":
        return self(
            id=data["short_id"],
            path=data["paths"][0],
            bytes=data["summary"]["total_bytes_processed"],
            backup_time=datetime.strptime(data["time"][:26], "%Y-%m-%dT%H:%M:%S.%f"),
        )


class VolumeMgr(object):
    cfg: MainConfig

    def __init__(self, cfg: MainConfig):
        self.cfg = cfg

    def volume_list(self, label_filters: KV = {}) -> List[KV]:
        # get the volume list that created by sandock
        volume_list_cmd = [
            self.cfg.execution.docker_bin,
            "volume",
            "ls",
            "--format=json",
        ]

        volume_list_cmd.extend(
            [f"--filter=label={k}='{v}'" for k, v in label_filters.items()]
        )

        exec_result = run_shell(command=volume_list_cmd)
        return [json.loads(line) for line in exec_result.stdout.split()]

    @property
    def created_by_sandock(self) -> List[KV]:
        # get the volume list that created by sandock

        return self.volume_list(label_filters={VOL_LABEL_CREATED_BY: "true"})

    def file_exists_in_vol(
        self, name: str, path: str, container: str = "docker.io/library/ubuntu:22.04"
    ) -> bool:
        # check if the mentioned path exists in specific volume ?
        check_file_cmd = [
            self.cfg.execution.docker_bin,
            "run",
            "-it",
            "--rm",
            "--entrypoint=test",
            f"-v {name}:/check_vol",
            container,
            "-f",
            f"/check_vol/{path}",
        ]
        result = run_shell(command=check_file_cmd, check_err=False)
        return result.returncode == 0

    def vol_exists(self, name: str) -> bool:
        # check if the mentioned volume exists ?
        vol_inspect_cmd = [
            self.cfg.execution.docker_bin,
            "volume",
            "inspect",
            f"'{name}'",
            "--format=json",
        ]

        cmd_out = run_shell(command=vol_inspect_cmd, check_err=False)

        if cmd_out.returncode == 0:
            vol_info = json.loads(cmd_out.stdout)
            log.debug(f"[vol] exists, info ~> {vol_info}")
            return True

        log.debug(f"vol by name {name} is not exists, err: {cmd_out.stderr}")
        return False

    @property
    def backup(self) -> "BackupMgr":
        return BackupMgr(vol_mgr=self)


class BackupMgr(object):
    vol_mgr: VolumeMgr
    # this only used once
    _backup_password: Optional[str] = None

    def __init__(self, vol_mgr: VolumeMgr):
        self.vol_mgr = vol_mgr
        if not self.cfg.no_password:
            self._backup_password = getpass("backup password: ")

    @property
    def cfg(self) -> BackupCfg:
        # backup related configuration
        return self.vol_mgr.cfg.backup

    @property
    def backup_dir(self) -> str:
        # return backup directory, escaping any home directory pattern
        return ensure_home_dir_special_prefix(path=self.cfg.path.default)

    @property
    def backup_config(self) -> Path:
        # returning restic config file inside repository, as the key indicator for some action (check initialized, etc)
        return Path(self.backup_dir, "config")

    def restic_run_cmd(
        self, extra_docker_params: List[str] = [], restic_args: List[str] = []
    ) -> List[str]:
        # compose restric command execution in container
        cmd = (
            [
                self.vol_mgr.cfg.execution.docker_bin,
                "run",
                "-it",
                "--rm",
                "--hostname=sandock",
                "--entrypoint=restic",
                f"-v {self.backup_dir}:{BACKUP_MOUNT}",
            ]
            + extra_docker_params
            + (
                [f"-e RESTIC_PASSWORD='{self._backup_password}'"]
                if self._backup_password is not None
                else []
            )
            + [self.cfg.restic.image]
            + [
                f"--repo={BACKUP_MOUNT}",
                f"--compression={self.cfg.restic.compression}",
                "--no-cache",
            ]
            + (["--insecure-no-password"] if self.cfg.no_password else [])
            + self.cfg.restic.extra_args
            + restic_args
        )

        return cmd

    def ensure_restic_repository(self) -> None:
        # ensuring restic repository (backup folder) has been initialized
        if self.backup_config.exists():
            log.debug(
                f"backup config binary is exists ({self.backup_config}), means it already initialized"
            )
            return

        log.warning(
            f"backup repository ({self.backup_dir}) is not initialized, running initialization"
        )

        init_repo_cmd = self.restic_run_cmd(restic_args=["init"])
        run_shell(command=init_repo_cmd, capture_output=False)

    def to_backup_path(self, vol_name: str) -> str:
        # returning the form of path name inside restic container
        return f"{SOURCE_MOUNT_PREFIX}{vol_name}"

    def create(self, targets: List[str], excludes: List[str] = []) -> None:
        # backup one or more volume target
        self.ensure_restic_repository()
        for target_vol in targets:
            if target_vol in self.cfg.volume_excludes:
                log.warning(
                    f"skipping backup for volume `{target_vol}`, listed in config exclude list"
                )
                continue

            if target_vol in excludes:
                log.warning(
                    f"skipping backup for volume `{target_vol}`, listed explicitly"
                )
                continue

            log.info(f"begin volume backup `{target_vol}`")
            backup_ignore_exists = False
            if self.vol_mgr.file_exists_in_vol(
                name=target_vol,
                path=BACKUP_IGNORE_REF_FILENAME,
                container=self.cfg.restic.image,
            ):
                log.debug(
                    f"backup ignore file ({BACKUP_IGNORE_REF_FILENAME}) in volume {target_vol}"
                )
                backup_ignore_exists = True

            source_mountpoint = self.to_backup_path(vol_name=target_vol)
            backup_cmd = self.restic_run_cmd(
                extra_docker_params=[f"-v {target_vol}:{source_mountpoint}:ro"],
                restic_args=[
                    "backup",
                    (
                        "--skip-if-unchanged"
                        if self.cfg.restic.no_snapshot_unless_changed
                        else ""
                    ),
                    "--group-by=paths",
                    (
                        f"--exclude-file={source_mountpoint}/{BACKUP_IGNORE_REF_FILENAME}"
                        if backup_ignore_exists
                        else ""
                    ),
                    source_mountpoint,
                ],
            )

            run_shell(command=backup_cmd, capture_output=False)

    def snapshot_list(
        self, specific_volname: Optional[str] = None, show_all: bool = False
    ) -> List[BackupSnapshot]:
        if not self.backup_config.exists():
            raise SandboxVolumeExec(
                f"backup repository ({self.backup_dir}) is not initialized"
            )

        snapshot_list_cmd = self.restic_run_cmd(restic_args=["snapshots", "--json"])
        if specific_volname:
            snapshot_list_cmd.append(
                f"--path={self.to_backup_path(vol_name=specific_volname)}"
            )

        if not show_all:
            snapshot_list_cmd.append("--latest=1")

        snapshots = json.loads(run_shell(command=snapshot_list_cmd).stdout)
        return [BackupSnapshot.from_raw(data=s) for s in snapshots]

    def get_snapshot_by(self, id: str) -> BackupSnapshot:
        snapshot_byid_cmd = self.restic_run_cmd(restic_args=["snapshots", id, "--json"])
        snapshot_byid_output = run_shell(command=snapshot_byid_cmd).stdout
        if "no matching ID found for prefix" in snapshot_byid_output:
            log.debug(f"snapshot by ID `{id}` is not found ~> {snapshot_byid_output}")
            raise SandboxVolumeNotFound(f"backup snapshot by id `{id}` is not found")

        return BackupSnapshot.from_raw(data=json.loads(snapshot_byid_output)[0])

    def restore(
        self,
        snapshot: Union[str, BackupSnapshot],
        target_volume: str,
        force: bool = False,
        excludes: List[str] = [],
        overwrite: Optional[str] = None,
    ) -> None:
        # restore backup result (snapshot) to a specific volume
        if not force and self.vol_mgr.vol_exists(name=target_volume):
            raise SandboxVolumeExec(
                f"volume by name `{target_volume}` is already exists, try to set force parameter to enforce it"
            )

        if isinstance(snapshot, str):
            snapshot = self.get_snapshot_by(id=snapshot)

        restore_mount_point = f"{RESTORE_MOUNT_PREFIX}{target_volume}"
        restore_cmd = self.restic_run_cmd(
            extra_docker_params=[f"-v {target_volume}:{restore_mount_point}"],
            restic_args=[
                "restore",
                f"{snapshot.id}:{snapshot.path}",
                f"--target={restore_mount_point}",
            ],
        )
        restore_cmd.extend([f"--exclude={ex}" for ex in excludes])

        if overwrite:
            restore_cmd.append(f"--overwrite={overwrite}")

        # TODO: check restore result
        run_shell(command=restore_cmd)

        log.info(f"backup snapshot by id {snapshot.id} ({snapshot.vol_name}) restored")



import os
from dataclasses import dataclass, field
from typing import Dict, Optional
from pathlib import Path
from ..shared import dict_merge, log, KV, CONFIG_PATH_ENV
from ..exceptions import SandboxExecConfig
from .image import ImageBuild
from .program import Program
from .config import Configuration
from .backup import Backup
from ._helpers import read_config, build_if_set, dot_config_finder


@dataclass
class Volume(object):
    driver: str = "local"
    driver_opts: Dict[str, str] = field(default_factory=dict)
    labels: Dict[str, str] = field(default_factory=dict)


@dataclass
class Network(object):
    driver: str = "bridge"
    driver_opts: Dict[str, str] = field(default_factory=dict)
    params: Dict[str, str] = field(default_factory=dict)


@dataclass
class Execution(object):
    docker_bin: str = "docker"
    container_name_prefix: str = "sandock-"
    property_override_prefix_arg: str = "sandbox-arg-"
    alias_program_prefix: str = ""


@dataclass
class MainConfig(object):
    execution: Execution = field(default_factory=Execution)
    config: Configuration = field(default_factory=Configuration)
    backup: Backup = field(default_factory=Backup)
    programs: Dict[str, Program] = field(default_factory=dict)
    volumes: Dict[str, Volume] = field(default_factory=dict)
    images: Dict[str, ImageBuild] = field(default_factory=dict)
    networks: Dict[str, Network] = field(default_factory=dict)

    def __post_init__(self) -> None:
        build_if_set(self, attr="config", cls=Configuration)
        expanded_configs = self.config.expand_configs()

        build_if_set(self, attr="execution", cls=Execution)
        build_if_set(self, attr="backup", cls=Backup)

        # configuration that use kv format if the value set as dict
        cls_mapper = dict(
            programs=Program, volumes=Volume, networks=Network, images=ImageBuild
        )

        for name, prop_cls in cls_mapper.items():
            prop_val = getattr(self, name)

            # expand if it's included
            expand_config = expanded_configs.get(name, {})
            if expand_config:
                log.debug(f"[config] expanding config attr {name} ~> {expand_config}")
                prop_val = dict_merge(prop_val, expand_config)

            for k, v in prop_val.items():
                if not isinstance(v, dict):
                    continue

                # extend to another declaration, the direct declaration will be the top priority
                extends: KV = {}
                for extend_key in v.pop("extends", []):
                    extend_props = prop_val.get(extend_key)
                    if not extend_props:
                        raise SandboxExecConfig(
                            f"no config found to be extended by key `{extend_key}`"
                        )

                    # if it's already as object, take the raw config that still in KV
                    if hasattr(extend_props, "_raw"):
                        extend_props = extend_props._raw

                    extends = dict_merge(extends, extend_props)

                if extends:
                    v = dict_merge(extends, v)

                config_obj = prop_cls(**v)
                config_obj._raw = v
                getattr(self, name)[k] = config_obj

        # at least need to define one program
        if not self.programs:
            raise ValueError("no program configured")


def load_config_file(path: str) -> MainConfig:
    """
    a thin wrapper for read configuration file to MainConfig object
    """

    return MainConfig(**read_config(path=path))


def main_config_finder(explicit_mention: Optional[str] = None) -> Optional[str]:
    """
    logic in finding configuration file by it's order
    """
    if explicit_mention:
        return explicit_mention

    env_conf = os.environ.get(CONFIG_PATH_ENV, None)
    if env_conf:
        return env_conf

    # dot config check
    dot_config = None
    home_dir = Path.home()
    home_dir_conf = dot_config_finder(directory=home_dir)
    if home_dir_conf:
        dot_config = home_dir_conf

    # last try for current directory
    current_dir = Path.cwd()
    if not dot_config and home_dir != current_dir:
        dot_config = dot_config_finder(directory=current_dir)

    return str(dot_config) if dot_config else None
